#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test websockets
"""

# doesn't work

# import socket, threading, time

# def handle(s):
#     print repr(s.recv(4096))
#     s.send('''
# HTTP/1.1 101 Web Socket Protocol Handshake\r
# Upgrade: WebSocket\r
# Connection: Upgrade\r
# WebSocket-Origin: http://localhost:8888\r
# WebSocket-Location: ws://localhost:9876/\r
# WebSocket-Protocol: sample
#   '''.strip() + '\r\n\r\n')
#     time.sleep(1)
#     s.send('\x00hello\xff')
#     time.sleep(1)
#     s.send('\x00world\xff')
#     s.close()

# s = socket.socket()
# s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
# s.bind(('', 9876))
# s.listen(1)
# while True:
#     t, _ = s.accept()
#     threading.Thread(target=handle, args=(t,)).start()

import cherrypy
from ws4py.server.cherrypyserver import WebSocketPlugin, WebSocketTool
from ws4py.websocket import WebSocket
import threading
import time
import json
cherrypy.config.update({'server.socket_port': 9000})
WebSocketPlugin(cherrypy.engine).subscribe()
cherrypy.tools.websocket = WebSocketTool()

idx = """
<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Web Socket Example</title>
    <meta charset="UTF-8">
    <script>
      window.onload = function() {
        var s = new WebSocket("ws://localhost:9876/");
        s.onopen = function(e) { alert("opened"); }
        s.onclose = function(e) { alert("closed"); }
        s.onmessage = function(evt) { $('#msg').append('<p>'+evt.data+'</p>');
                }; 
      };
    </script>
  </head>
    <body onLoad="pageScroll()">
      <div id="holder" style="width:600px; height:300px"></div>
    </body>
</html>
""".strip()

idx2 = """
<!DOCTYPE html>
<html>
<head>
<script src='http://ajax.googleapis.com/ajax/libs/jquery/1.8.3/jquery.min.js'></script>
<script src='https://raw.githubusercontent.com/flot/flot/master/jquery.flot.js'></script>
<script src='https://raw.githubusercontent.com/flot/flot/master/jquery.flot.time.js'></script>
<script>
    $(document).ready(function() {
            function debug(str) {
                $('#debug').append('<p>'+str+'</p>');
                };
            ws = new WebSocket('ws://localhost:9000/ws');

            var data = [];
            var qual = [];
            var count = 0;
    var plot = $.plot("#debug", [ data ], {
			series: {
				shadowSize: 0	// Drawing is faster without shadows
			},
			yaxis: {
				min: -1,
				max: 102
			},
			xaxis: {
				show: true,
                mode: "time"
                                
			}
		});
    

/* Define websocket handlers */
      ws.onmessage = function(evt) {
                var d = JSON.parse(evt.data);
                //$('#msg').append(evt.data+'<br />');
                //$('#msg').append(d.mtype +'<br />');
                if(d.mtype=="have"){
                    //data = data.slice(1);
                    data.push([d.timecode, parseFloat(d.elevation)]);
                    qual.push([d.timecode, parseFloat(d.quality)]);
                    //$('#msg').append(parseFloat(d.elevation) +'<br />');
                    //plot.setData([ data, qual ]);
                    //plot.draw();
                    $.plot("#debug", [ data, qual ], {
			series: {
				shadowSize: 0	// Drawing is faster without shadows
			},
			yaxis: {
				min: -1,
				max: 102
			},
			xaxis: {
				show: true,
                mode: "time"
                                
			}
		});
                    count = count + 1;
                }
                document.scrollTop = document.scrollHeight;
                };
      ws.onclose = function() {
                debug('socket closed');
                };
      ws.onopen = function() {
                debug('connected...');
                ws.send('Hello.\\n');
            };
            });
/* Define click handler */

    $(document).click(function () {
                ws.send('Click!');
            });
</script>
</head>
<body>
    <div id='debug' style="width:600px;height:300px"></div>
    <div id='plots'></div>
    <div id='msg'></div>
</body>
</html>
""".strip()


class Root(object):

    @cherrypy.expose
    def index(self):
        # return 'some HTML with a websocket javascript connection'
        return idx2

    @cherrypy.expose
    @cherrypy.tools.websocket(on=False)
    def ws(self):
        # you can access the class instance through
        handler = cherrypy.request.ws_handler
        # return handler

item = Root()

from ws4py.messaging import TextMessage
from posttroll.subscriber import Subscribe

# cherrypy.config.update({'/ws': {'tools.websocket.on': True,
#                                'tools.websocket.handler_cls': WebSocket}})
cherrypy.tree.mount(Root(), "/", config={'/ws': {'tools.websocket.on': True,
                                                 'tools.websocket.handler_cls': WebSocket}})

# cherrypy.quickstart(item, '/', config={'/ws': {'tools.websocket.on': True,
#                                               'tools.websocket.handler_cls': EchoWebSocket}})
# pause
cherrypy.engine.start()


def send(msg):
    cherrypy.engine.publish("websocket-broadcast", TextMessage(msg))

from trollcast.client import create_subscriber
import sys
import calendar
try:
    sub = create_subscriber(sys.argv[1])
    for message in sub.recv(1):
        if message is None:
            continue
        elif message.type == "have":
            data = message.data
            data["mtype"] = "have"
            data["sender"] = message.sender
            #data["timecode"] = data["timecode"].isoformat()
            data["timecode"] = calendar.timegm(
                data["timecode"].timetuple()) * 1000
            send(json.dumps(data))
        elif message.type == "heartbeat":
            print "heartbeat"
            data = message.data
            data["mtype"] = "heartbeat"
            data["sender"] = message.sender
            send(json.dumps(data))

    # with Subscribe(services=[""], addr_listener=True) as sub:
    #     for msg in sub.recv(1):
    #         if msg:
    #             if msg.type in ["log.debug", "log.info",
    #                             "log.warning", "log.error",
    #                             "log.critical"]:
    #                 send(msg.type + " " + msg.subject + " " +
    #                      msg.sender + " " +
    #                      str(msg.data) + " " +
    #                      str(msg.time))

    #             elif msg.binary:
    #                 send(msg.subject + " " +
    #                      msg.sender + " " +
    #                      msg.type + " " +
    #                      "[binary] " +
    #                      str(msg.time))
    #             else:
    #                 send(msg.subject + " " +
    #                      msg.sender + " " +
    #                      msg.type + " " +
    #                      str(msg.data) + " " +
    #                      str(msg.time))

finally:
    cherrypy.engine.stop()

pause
#cherrypy.engine.publish("websocket-broadcast", msg)


class EchoWebSocket(WebSocket):

    def received_message(self, message):
        """
Automatically sends back the provided ``message`` to
its originating endpoint.
"""
        self.send(message.data, message.is_binary)
        # cnt = 0
        # while True:

        #     msg = TextMessage("hej " + str(cnt))
        #     self.send(msg.data, msg.is_binary)
        #     time.sleep(1)
        #     cnt += 1

        with Subscribe(services=[""], addr_listener=True) as sub:
            for msg in sub.recv(1):
                if msg:
                    if msg.type in ["log.debug", "log.info",
                                    "log.warning", "log.error",
                                    "log.critical"]:
                        self.send(msg.type + " " + msg.subject + " " +
                                  msg.sender + " " +
                                  str(msg.data) + " " +
                                  str(msg.time))

                    elif msg.binary:
                        self.send(msg.subject + " " +
                                  msg.sender + " " +
                                  msg.type + " " +
                                  "[binary] " +
                                  str(msg.time))
                    else:
                        self.send(msg.subject + " " +
                                  msg.sender + " " +
                                  msg.type + " " +
                                  str(msg.data) + " " +
                                  str(msg.time))
#                if not self.loop:
#                    self.send("Stop logging")
#                    break

# cherrypy.quickstart(item, '/', config={'/ws': {'tools.websocket.on': True,
#                                                 'tools.websocket.handler_cls': EchoWebSocket}})


class listener(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self.cpy = None

    def run(self):
        #self.cpy = cherrypy.quickstart( self )
        self.cpy = cherrypy.quickstart(item, '/', config={'/ws': {'tools.websocket.on': True,
                                                                  'tools.websocket.handler_cls': EchoWebSocket}})

    def stop(self):
        self.cpy.stop()


interf = listener()
interf.start()

print "ready"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2013, 2014, 2015 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Trollcast client. Leeches all it can :)
todo:
- connection between institutes is shutdown after a while (2 hours ?)
- filename are wrong (1 year to old)
- Option for new log file every day? Now log files are quite big after few days.
- resets connection to mirror in case of timeout.

"""
from __future__ import with_statement

import logging
from ConfigParser import ConfigParser, NoOptionError
from Queue import Queue, Empty
from datetime import timedelta, datetime
from threading import Thread, Timer, Event, Lock
from urlparse import urlsplit
import numpy as np
from posttroll.message import Message, strp_isoformat
from posttroll import context
from zmq import Context, REQ, LINGER, Poller, POLLIN, SUB, SUBSCRIBE, zmq_version
import os.path
from urlparse import urlunparse
from trollsift import compose
import warnings

logger = logging.getLogger(__name__)

# TODO: what if a scanline never arrives ? do we wait for it forever ?
# TODO: subscriber reset

# FIXME: this should be configurable depending on the type of data.
LINES_PER_SECOND = 6
LINE_SIZE = 11090 * 2
CLIENT_TIMEOUT = timedelta(seconds=45)

REQ_TIMEOUT = 1000
if zmq_version().startswith("2."):
    REQ_TIMEOUT *= 1000
BUFFER_TIME = 2.0


class Subscriber(object):

    def __init__(self, addresses, translate=False):
        self._addresses = addresses
        self._translate = translate
        self.subscribers = []
        self._poller = Poller()
        for addr in self._addresses:
            subscriber = context.socket(SUB)
            subscriber.setsockopt(SUBSCRIBE, "pytroll")
            subscriber.connect(addr)
            self.subscribers.append(subscriber)
            self._poller.register(subscriber)
        self._lock = Lock()
        self._loop = True

    @property
    def sub_addr(self):
        return dict(zip(self.subscribers, self._addresses))

    @property
    def addr_sub(self):
        return dict(zip(self._addresses, self.subscribers))

    def reset(self, addr):
        with self._lock:
            idx = self._addresses.index(addr)
            self._poller.unregister(self.subscribers[idx])
            self.subscribers[idx].setsockopt(LINGER, 0)
            self.subscribers[idx].close()
            self.subscribers[idx] = context.socket(SUB)
            self.subscribers[idx].setsockopt(SUBSCRIBE, "pytroll")
            self.subscribers[idx].connect(addr)
            self._poller.register(self.subscribers[idx], POLLIN)

    def recv(self, timeout=None):
        """Receive a message, timeout in seconds.
        """

        if timeout:
            timeout *= 1000

        while(self._loop):
            with self._lock:
                if not self._loop:
                    break
                subs = dict(self._poller.poll(timeout=timeout))
                if subs:
                    for sub in self.subscribers:
                        if sub in subs and subs[sub] == POLLIN:
                            msg = Message.decode(sub.recv())
                            if self._translate:
                                url = urlsplit(self.sub_addr[sub])
                                host = url[1].split(":")[0]
                                msg.sender = (msg.sender.split("@")[0]
                                              + "@" + host)
                            yield msg
                else:
                    yield None

    def stop(self):
        """Stop the subscriber
        """
        with self._lock:
            self._loop = False
            for sub in self.subscribers:
                self._poller.unregister(sub)
                sub.setsockopt(LINGER, 0)
                sub.close()


def create_subscriber(cfgfile):
    """Create a new subscriber for all the remote hosts in cfgfile.
    """

    cfg = ConfigParser()
    cfg.read(cfgfile)
    addrs = []
    for host in cfg.get("local_reception", "remotehosts").split():
        addrs.append("tcp://" +
                     cfg.get(host, "hostname") + ":" + cfg.get(host, "pubport"))
    localhost = cfg.get("local_reception", "localhost")
    addrs.append("tcp://" +
                 cfg.get(localhost, "hostname") + ":" +
                 cfg.get(localhost, "pubport"))
    logger.debug("Subscribing to " + str(addrs))
    return Subscriber(addrs, translate=True)


class RTimer(Thread):

    def __init__(self, tries, warning_message, function, *args, **kwargs):
        Thread.__init__(self)
        self.event = Event()
        self.interval = 40
        self.loop = True
        self.tries = tries
        self.attempt = 0
        self.args = args
        self.kwargs = kwargs
        self.alert_func = function
        self.warning = warning_message

    def reset(self):
        self.event.set()
        self.attempt = 0

    def alert(self):
        self.attempt += 1
        if self.attempt >= self.tries:
            logger.warning(self.warning)
            self.alert_func(*self.args, **self.kwargs)
            self.reset()
        else:
            logger.warning(self.warning)

    def run(self):
        while self.loop:
            self.event.wait(self.interval)
            if not self.event.is_set() and self.loop:
                self.alert()
            elif self.event.is_set():
                self.event.clear()

    def stop(self):
        self.loop = False
        self.event.set()


def reset_subscriber(subscriber, addr):
    logger.warning("Resetting connection to " + addr)
    subscriber.reset(addr)


def create_timers(cfgfile, subscriber):
    cfg = ConfigParser()
    cfg.read(cfgfile)
    addrs = []
    timers = {}
    for host in cfg.get("local_reception", "remotehosts").split():
        addrs.append("tcp://" +
                     cfg.get(host, "hostname") + ":" + cfg.get(host, "pubport"))
    localhost = cfg.get("local_reception", "localhost")
    addrs.append("tcp://" +
                 cfg.get(localhost, "hostname") + ":" +
                 cfg.get(localhost, "pubport"))
    for addr in addrs:
        timers[addr] = RTimer(1, addr + " seems to be down, no hearbeat received",
                              reset_subscriber, subscriber, addr)

        timers[addr].start()
    return timers


def create_requesters(cfgfile):
    """Create requesters to all the configure remote hosts.
    """
    cfg = ConfigParser()
    cfg.read(cfgfile)
    station = cfg.get("local_reception", "station")
    requesters = {}
    for host in cfg.get("local_reception", "remotehosts").split():
        pubport = cfg.get(host, "pubport")
        host, port = (cfg.get(host, "hostname"), cfg.get(host, "reqport"))
        requesters[host] = Requester(host, port, station, pubport)
    host = cfg.get("local_reception", "localhost")
    pubport = cfg.get(host, "pubport")
    host, port = (cfg.get(host, "hostname"), cfg.get(host, "reqport"))
    url = "tcp://" + host + ":" + port
    requesters[host] = Requester(host, port, station, pubport)
    return requesters


class SimpleRequester(object):

    """Base requester class.
    """

    request_retries = 3

    def __init__(self, host, port):
        self._socket = None
        self._reqaddress = "tcp://" + host + ":" + str(port)
        self._poller = Poller()
        self._lock = Lock()
        self.failures = 0
        self.jammed = False

        self.connect()

    def connect(self):
        """Connect to the server
        """
        self._socket = context.socket(REQ)
        self._socket.connect(self._reqaddress)
        self._poller.register(self._socket, POLLIN)

    def stop(self):
        """Close the connection to the server
        """
        self._socket.setsockopt(LINGER, 0)
        self._socket.close()
        self._poller.unregister(self._socket)

    def reset_connection(self):
        """Reset the socket
        """
        self.stop()
        self.connect()

    def __del__(self, *args, **kwargs):
        self.stop()

    def send_and_recv(self, msg, timeout=REQ_TIMEOUT):

        logger.debug("Locking and requesting: " + str(msg))
        with self._lock:
            retries_left = self.request_retries
            request = str(msg)
            self._socket.send(request)
            rep = None
            while retries_left:
                socks = dict(self._poller.poll(timeout))
                if socks.get(self._socket) == POLLIN:
                    reply = self._socket.recv()
                    if not reply:
                        logger.error("Empty reply!")
                        break
                    rep = Message(rawstr=reply)
                    if rep.binary:
                        logger.debug("Got reply: "
                                     + " ".join(str(rep).split()[:6]))
                    else:
                        logger.debug("Got reply: " + str(rep))
                    self.failures = 0
                    self.jammed = False
                    break
                else:
                    logger.warning("Timeout from " + str(self._reqaddress)
                                   + ", retrying...")
                    # Socket is confused. Close and remove it.
                    self.stop()
                    retries_left -= 1
                    if retries_left <= 0:
                        logger.error("Server doesn't answer, abandoning... " +
                                     str(self._reqaddress))
                        self.connect()
                        self.failures += 1
                        if self.failures == 5:
                            logger.critical("Server jammed ? %s",
                                            self._reqaddress)
                            self.jammed = True
                        break
                    logger.info("Reconnecting and resending " + str(msg))
                    # Create new connection
                    self.connect()
                    self._socket.send(request)
        logger.debug("Release request lock")
        return rep


class Requester(SimpleRequester):

    """Make a request connection, waiting to get scanlines .
    """

    def __init__(self, host, port, station, pubport=None):
        SimpleRequester.__init__(self, host, port)
        self._station = station
        self._pubport = pubport
        self.blocked = False

    def send(self, msg):
        """Send a message.
        """
        warnings.warn("Send method of Requester is deprecated",
                      DeprecationWarning)
        return self._socket.send(str(msg))

    def recv(self, timeout=None):
        """Receive a message. *timeout* in ms.
        """
        warnings.warn("Recv method of Requester is deprecated",
                      DeprecationWarning)
        if self._poller.poll(timeout):
            self.blocked = False
            return Message(rawstr=self._socket.recv())
        else:
            raise IOError("Timeout from " + str(self._reqaddress))

    def ping(self):
        """Send a ping.
        """

        msg = Message('/oper/polar/direct_readout/' + self._station,
                      'ping',
                      {"station": self._station})
        tic = datetime.now()
        response = self.send_and_recv(msg, REQ_TIMEOUT)
        toc = datetime.now()
        if response is None:
            return None, toc - tic
        else:
            return response.data["station"], toc - tic

    def get_line(self, satellite, utctime):
        """Get the scanline of *satellite* at *utctime*.
        """
        msg = Message('/oper/polar/direct_readout/' + self._station,
                      'request',
                      {"type": "scanline",
                       "satellite": satellite,
                       "utctime": utctime.isoformat()})

        try:
            resp = self.send_and_recv(msg, REQ_TIMEOUT)
            if resp.type == "missing":
                return None
            else:
                return resp.data
        except AttributeError:
            return None

    def get_slice(self, satellite, start_time, end_time):
        """Get a slice of scanlines.
        """
        msg = Message('/oper/polar/direct_readout/' + self._station,
                      'request',
                      {"type": 'scanlines',
                       "satellite": satellite,
                       "start_time": start_time.isoformat(),
                       "end_time": end_time.isoformat()})
        self.send(msg)
        return self.recv(REQ_TIMEOUT).data

    def send_lineinfo(self, sat, utctime, elevation, filename, pos):
        """Send information to our own server.
        """

        msg = Message('/oper/polar/direct_readout/' + self._station,
                      'notice',
                      {"type": 'scanline',
                       "satellite": sat,
                       "utctime": utctime.isoformat(),
                       "elevation": elevation,
                       "filename": filename,
                       "file_position": pos})
        response = self.send_and_recv(msg, REQ_TIMEOUT)
        return response


def create_publisher(cfgfile):
    cfg = ConfigParser()
    cfg.read(cfgfile)
    try:
        publisher = cfg.get("local_reception", "publisher")
    except NoOptionError:
        return None
    if publisher:
        from posttroll.publisher import NoisyPublisher
        publisher = NoisyPublisher(publisher, 0)
        publisher.start()
        return publisher


class HaveBuffer(Thread):

    """Listen to incomming have messages.
    """

    def __init__(self, cfgfile="sattorrent.cfg"):
        Thread.__init__(self)
        self._sub = create_subscriber(cfgfile)
        self._hb = create_timers(cfgfile, self._sub)
        self._publisher = create_publisher(cfgfile)
        cfg = ConfigParser()
        cfg.read(cfgfile)
        try:
            self._out = cfg.get("local_reception", "output_file")
        except NoOptionError:
            self._out = "./{utctime:%Y%m%d%H%M%S}_{platform:4s}_{number:2s}.trollcast.hmf"
        localhost = cfg.get("local_reception", "localhost")
        self._hostname = cfg.get(localhost, "hostname")
        self._station = cfg.get("local_reception", "station")
        self.scanlines = {}
        self._queues = []
        self._requesters = {}
        self._timers = {}
        self._ping_times = {}

    def add_queue(self, queue):
        """Adds a queue to dispatch have messages to
        """
        self._queues.append(queue)

    def del_queue(self, queue):
        """Deletes a dispatch queue.
        """
        self._queues.remove(queue)

    def send_to_queues(self, sat, utctime):
        """Send scanline at *utctime* to queues.
        """
        try:
            self._timers[(sat, utctime)].cancel()
            del self._timers[(sat, utctime)]
        except KeyError:
            pass
        for queue in self._queues:
            queue.put_nowait((sat, utctime, self.scanlines[sat][utctime]))

    def run(self):

        for message in self._sub.recv(1):
            if message is None:
                continue
            if(message.type == "have"):
                sat = message.data["satellite"]
                utctime = strp_isoformat(message.data["timecode"])
                # This should take care of address translation.
                senderhost = message.sender.split("@")[1]
                sender = (senderhost + ":" +
                          message.data["origin"].split(":")[1])
                elevation = message.data["elevation"]
                try:
                    quality = message.data["quality"]
                except KeyError:
                    quality = 100
                self.scanlines.setdefault(sat, {})
                if utctime not in self.scanlines[sat]:
                    self.scanlines[sat][utctime] = [
                        (sender, elevation, quality, self._ping_times.get(senderhost, 3600000))]
                    # TODO: This implies that we always wait BUFFER_TIME before
                    # sending to queue. In the case were the "have" messages of
                    # all servers were sent in less time, we should not be
                    # waiting...
                    if len(self._requesters) == 1:
                        self.send_to_queues(sat, utctime)
                    else:
                        timer = Timer(BUFFER_TIME,
                                      self.send_to_queues,
                                      args=[sat, utctime])
                        timer.start()
                        self._timers[(sat, utctime)] = timer
                else:
                    # Since append is atomic in CPython, this should work.
                    # However, if it is not, then this is not thread safe.
                    self.scanlines[sat][utctime].append((sender, elevation,
                                                         quality, self._ping_times.get(senderhost, 3600000)))
                    if (len(self.scanlines[sat][utctime]) ==
                            len(self._requesters)):
                        self.send_to_queues(sat, utctime)
            elif(message.type == "heartbeat"):
                senderhost = message.sender.split("@")[1]
                sender = ("tcp://" + senderhost +
                          ":" + message.data["addr"].split(":")[1])
                logger.debug("receive heartbeat from " + str(sender) +
                             ": " + str(message))
                self._hb[str(sender)].reset()

                for addr, req in self._requesters.items():
                    # can we get the ip adress from the socket somehow ?
                    # because right now the pubaddr and sender are not the same
                    # (name vs ip)
                    if addr == senderhost:
                        rstation, rtime = req.ping()
                        if rstation is None:
                            logger.warning("Can't ping " + str(sender))
                        else:
                            rtime = (rtime.seconds * 1000 +
                                     rtime.microseconds / 1000.0)
                            self._ping_times[senderhost] = rtime
                            logger.debug("ping roundtrip to " + rstation +
                                         ": " + str(rtime) +
                                         "ms")
                        break

    def stop(self):
        """Stop buffering.
        """
        self._sub.stop()
        for timer in self._hb.values():
            timer.stop()
        if self._publisher:
            self._publisher.stop()


def compute_line_times(utctime, start_time, end_time):
    """Compute the times of lines if a swath order depending on a reference
    *utctime*.
    """
    offsets = (np.arange(0, 1, 1.0 / LINES_PER_SECOND) +
               utctime.microsecond / 1000000.0)
    offsets = offsets.round(3)
    offsets[offsets > 1] -= 1
    offsets -= start_time.microsecond
    offset = timedelta(seconds=min(abs(offsets)))
    time_diff = (end_time - start_time - offset)
    time_diff = time_diff.seconds + time_diff.microseconds / 1000000.0
    nblines = int(np.ceil(time_diff * LINES_PER_SECOND))
    rst = start_time + offset
    linepos = [rst + timedelta(seconds=round(i * 1.0 /
                                             LINES_PER_SECOND, 3))
               for i in range(nblines)]
    linepos = set(linepos)
    return linepos


true_names = {"NOAA 19": "NOAA-19",
              "NOAA 18": "NOAA-18",
              "NOAA 15": "NOAA-15"}


class Client(HaveBuffer):

    """The client class.
    """

    def __init__(self, cfgfile="sattorrent.cfg"):
        HaveBuffer.__init__(self, cfgfile)
        self._requesters = create_requesters(cfgfile)
        self.cfgfile = cfgfile
        self.loop = True

    def get_lines(self, satellite, scanline_dict):
        """Retrieve the best (highest elevation) lines of *scanline_dict*.
        """

        for utctime, hosts in scanline_dict.iteritems():
            hostname, elevation = max(hosts, key=(lambda x: x[1]))
            host = hostname.split(":")[0]

            logger.debug("requesting " + " ".join([str(satellite),
                                                   str(utctime),
                                                   str(host)]))
            data = self._requesters[host].get_line(satellite, utctime)

            yield utctime, data, elevation

    def get_all(self, satellites):
        """Retrieve all the available scanlines from the stream, and save them.
        """
        sat_last_seen = {}
        sat_lines = {}
        first_time = None
        for sat in satellites:
            sat_lines[sat] = {}
        queue = Queue()
        self.add_queue(queue)
        try:
            while self.loop:
                try:
                    sat, utctime, senders = queue.get(True, 2)
                    if sat not in satellites:
                        continue

                    if sat not in sat_last_seen:
                        logger.info("Start receiving data for " + sat)

                    logger.debug("Picking line " + " ".join([str(utctime),
                                                             str(senders)]))
                    # choose the highest quality, lowest ping time, highest elevation.
                    sender_elevation_quality = sorted(senders,
                                                      key=(lambda x: (x[2],
                                                                      -x[3],
                                                                      x[1])))
                    best_req = None
                    for sender, elevation, quality, ping_time in reversed(sender_elevation_quality):
                        best_req = self._requesters[sender.split(":")[0]]
                        if best_req.jammed:
                            continue
                        sat_last_seen[sat] = datetime.utcnow(), elevation
                        logger.debug("requesting " +
                                     " ".join([str(sat), str(utctime),
                                               str(sender), str(elevation)]))
                        # TODO: this should be parallelized, and timed. In case of
                        # TODO: Choking ?
                        line = best_req.get_line(sat, utctime)
                        if line is None:
                            logger.warning("Could not retrieve line %s",
                                           str(utctime))
                        else:
                            sat_lines[sat][utctime] = line
                            if first_time is None and quality == 100:
                                first_time = utctime
                            break

                    if best_req is None:
                        logger.debug("No working connection, could not retrieve"
                                     " line %s", str(utctime))
                        continue


                except Empty:
                    pass
                for sat, (utctime, elevation) in sat_last_seen.items():
                    if (utctime + CLIENT_TIMEOUT < datetime.utcnow() or
                        (utctime + timedelta(seconds=3) < datetime.utcnow() and
                         elevation < 0.5 and
                         len(sat_lines[sat]) > 100)):
                        # write the lines to file
                        try:
                            first_time = (first_time
                                          or min(sat_lines[sat].keys()))
                            last_time = max(sat_lines[sat].keys())
                            logger.info(sat +
                                        " seems to be inactive now, writing file.")
                            fdict = {}
                            fdict["platform"], fdict["number"] = sat.split()
                            fdict["utctime"] = first_time
                            filename = compose(self._out, fdict)
                            with open(filename, "wb") as fp_:
                                for linetime in sorted(sat_lines[sat].keys()):
                                    fp_.write(sat_lines[sat][linetime])
                            if self._publisher:
                                to_send = {}
                                to_send["platform_name"] = true_names[sat]
                                to_send["format"] = "HRPT"
                                to_send["start_time"] = first_time
                                to_send["end_time"] = last_time
                                to_send["data_processing_level"] = "0"
                                to_send["uid"] = os.path.basename(
                                    filename)
                                fullname = os.path.realpath(filename)
                                to_send["uri"] = urlunparse(("ssh",
                                                             self._hostname,
                                                             fullname,
                                                             "",
                                                             "",
                                                             ""))
                                if sat == "NOAA 15":
                                    to_send["sensor"] = ("avhrr/3",
                                                         "amsu-a",
                                                         "amsu-b",
                                                         "hirs/3")
                                elif sat in ["NOAA 19", "NOAA 18"]:
                                    to_send["sensor"] = ("avhrr/3",
                                                         "mhs",
                                                         "amsu-a",
                                                         "hirs/4")

                                to_send["type"] = "binary"
                                msg = Message("/".join(
                                    ("",
                                     to_send["format"],
                                     to_send[
                                         "data_processing_level"],
                                     self._station)),
                                    "file",
                                    to_send)
                                logger.debug("publishing %s", str(msg))
                                self._publisher.send(str(msg))

                        except ValueError:
                            logger.info("Got no lines for " + sat)
                            continue
                        finally:
                            sat_lines[sat] = {}
                            del sat_last_seen[sat]
                            first_time = None
        except KeyboardInterrupt:
            for sat, (utctime, elevation) in sat_last_seen.items():
                logger.info(sat + ": writing file.")
                first_time = (first_time
                              or min(sat_lines[sat].keys()))
                filename = first_time.isoformat() + sat + ".hmf"
                with open(filename, "wb") as fp_:
                    for linetime in sorted(sat_lines[sat].keys()):
                        fp_.write(sat_lines[sat][linetime])

                sat_lines[sat] = {}
                del sat_last_seen[sat]
            raise

    def order(self, time_slice, satellite, filename):
        """Get all the scanlines for a *satellite* within a *time_slice* and
        save them in *filename*. The scanlines will be saved in a contiguous
        manner.
        """
        start_time = time_slice.start
        end_time = time_slice.stop

        saved = []

        # Create a file of the right length, filled with zeros. The alternative
        # would be to store all the scanlines in memory.
        tsize = (end_time - start_time).seconds * LINES_PER_SECOND * LINE_SIZE
        with open(filename, "wb") as fp_:
            fp_.write("\x00" * (tsize))

        # Do the retrieval.
        with open(filename, "r+b") as fp_:

            queue = Queue()
            self.add_queue(queue)

            linepos = None

            lines_to_get = {}

            # first, get the existing scanlines from self (client)
            logger.info("Getting list of existing scanlines from client.")
            for utctime, hosts in self.scanlines.get(satellite, {}).iteritems():
                if(utctime >= start_time and
                   utctime < end_time and
                   utctime not in saved):
                    lines_to_get[utctime] = hosts

            # then, get scanlines from the server
            logger.info("Getting list of existing scanlines from server.")
            for host, req in self._requesters.iteritems():
                try:
                    response = req.get_slice(satellite, start_time, end_time)
                    for utcstr, elevation in response:
                        utctime = strp_isoformat(utcstr)
                        lines_to_get.setdefault(utctime, []).append((host,
                                                                     elevation))
                except IOError, e__:
                    logger.warning(e__)

            # get lines with highest elevation and add them to current scene
            logger.info("Getting old scanlines.")
            for utctime, data, elevation in self.get_lines(satellite,
                                                           lines_to_get):
                if linepos is None:
                    linepos = compute_line_times(utctime, start_time, end_time)

                time_diff = utctime - start_time
                time_diff = (time_diff.seconds
                             + time_diff.microseconds / 1000000.0)
                pos = LINE_SIZE * int(np.floor(time_diff * LINES_PER_SECOND))
                fp_.seek(pos, 0)
                fp_.write(data)
                self.send_lineinfo_to_server(satellite, utctime, elevation,
                                             filename, pos)
                saved.append(utctime)
                linepos -= set([utctime])

            # then, get the newly arrived scanlines
            logger.info("Getting new scanlines")
            #timethres = datetime.utcnow() + CLIENT_TIMEOUT
            delay = timedelta(days=1000)
            timethres = datetime.utcnow() + delay
            while ((start_time > datetime.utcnow()
                    or timethres > datetime.utcnow())
                   and ((linepos is None) or (len(linepos) > 0))):
                try:
                    sat, utctime, senders = queue.get(True,
                                                      CLIENT_TIMEOUT.seconds)
                    logger.debug("Picking line " + " ".join([str(utctime),
                                                             str(senders)]))
                    # choose the highest elevation
                    sender, elevation = max(senders, key=(lambda x: x[1]))

                except Empty:
                    continue

                if linepos is None:
                    linepos = compute_line_times(utctime, start_time, end_time)

                if(sat == satellite and
                   utctime >= start_time and
                   utctime < end_time and
                   utctime not in saved):
                    saved.append(utctime)

                    # getting line
                    logger.debug("requesting " +
                                 " ".join([str(satellite), str(utctime),
                                           str(sender), str(elevation)]))
                    host = sender.split(":")[0]
                    # TODO: this should be parallelized, and timed. I case of
                    # failure, another source should be used. Choking ?
                    line = self._requesters[host].get_line(satellite, utctime)

                    # compute line position in file
                    time_diff = utctime - start_time
                    time_diff = (time_diff.seconds
                                 + time_diff.microseconds / 1000000.0)
                    pos = LINE_SIZE * int(np.floor(time_diff *
                                                   LINES_PER_SECOND))
                    fp_.seek(pos, 0)
                    fp_.write(line)
                    self.send_lineinfo_to_server(satellite, utctime, elevation,
                                                 filename, pos)
                    # removing from line check list
                    linepos -= set([utctime])

                    delay = min(delay, datetime.utcnow() - utctime)
                    if len(linepos) > 0:
                        timethres = max(linepos) + CLIENT_TIMEOUT + delay
                    else:
                        timethres = datetime.utcnow()

            # shut down
            self.del_queue(queue)

    def send_lineinfo_to_server(self, *args, **kwargs):
        """Send information to our own server.
        """
        cfg = ConfigParser()
        cfg.read(self.cfgfile)
        host = cfg.get("local_reception", "localhost")
        host, port = (cfg.get(host, "hostname"),  cfg.get(host, "reqport"))
        del port
        self._requesters[host].send_lineinfo(*args, **kwargs)

    def stop(self):
        HaveBuffer.stop(self)
        self.loop = False
        for req in self._requesters.values():
            req.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import requests
import os
auth = ("mraspaud", "esasucks")
#params = {'q': "+".join('productType:"GRD"  AND  swathIdentifier:"EW"  AND( ingestionDate:[2014-12-07T00:00:00.000Z TO 2014-12-08T23:59:59.999Z ] )  AND ( footprint:"Intersects(POLYGON ((6.1299197292328 53.42977963583,30.299841604233 53.42977963583,30.299841604233 66.120926899993,6.1299197292328 66.120926899993,6.1299197292328 53.42977963583)))" )'.split())}
params = {
    #'q': 'footprint:"Intersects(POLYGON((-4.53 29.85,26.75 29.85,26.75 46.80,-4.53 46.80,-4.53 29.85)))"'
    #'q': 'footprint:"Intersects(POLYGON ((6.29 61.61,24.54 67.40,31.48 60.43,26.78 54.46,6.12 50.86,6.29 61.61)))"'
    #'q': 'beginPosition:[NOW-3MONTHS TO NOW] AND endPosition:[NOW-3MONTHS TO NOW]'
    #'q': 'swathIdentifier:"EW" AND footprint:"Intersects(POLYGON ((6.1299197292328 53.42977963583,30.299841604233 53.42977963583,30.299841604233 66.120926899993,6.1299197292328 66.120926899993,6.1299197292328 53.42977963583)))"'
    'q': 'swathIdentifier:"EW"'
}


# r = requests.get(
#    'https://scihub.esa.int/dhus/search', params=params, auth=auth)
# print r.text


import subprocess as sp
proc = sp.Popen(["curl", "-gu", "mraspaud:esasucks",
                 #"https://scihub.esa.int/dhus/search?q=ingestiondate:[NOW-200DAYS+TO+NOW]+AND+producttype:GRD+AND+swathIdentifier:EW+AND+footprint:\"Intersects(POLYGON+((53.4300000000000%206.1300000000000,66.1200000000000%206.1300000000000,66.1200000000000%2030.3000000000000,53.4300000000000%2030.3000000000000,53.4300000000000%206.1300000000000%20))))\"&rows=10000&start=0"
                 #"https://scihub.esa.int/dhus/search?q=ingestiondate:[1970-12-05T16:33:18.525585299Z%20TO%20NOW]%20AND%20producttype:GRD+AND+(%20footprint:%22Intersects(POLYGON((53.43%206.13,66.12%206.13,66.12%2030.30,53.43%2030.30,53.43%206.13%20)))%22)&rows=10000&start=0"
                 "https://scihub.esa.int/dhus/search?q=ingestiondate:[NOW-7DAYS+TO+NOW]+AND+producttype:GRD+AND+swathIdentifier:\"EW\"+AND+footprint:\"Intersects(POLYGON+((6.13+53.43,6.13+66.15,30.30+66.12,30.30+53.43,6.13+53.43)))\"&rows=10000&start=0"
                 ], stdout=sp.PIPE, stderr=sp.PIPE)

text, errors = proc.communicate()

print text

import feedparser
from datetime import datetime, timedelta
d = feedparser.parse(text)

print d.feed.title
print len(d.entries)


def download_file(url, local_filename):
    # NOTE the stream=True parameter
    r = requests.get(url, stream=True, auth=auth)
    with open(local_filename, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                f.flush()
    return local_filename

for entry in d.entries:
    date = datetime.strptime(entry['summary'].split(',')[0].split()[1],
                             "%Y-%m-%dT%H:%M:%S.%fZ")
    if date > datetime.utcnow() - timedelta(days=2):
        fname = os.path.join(
            "/data/temp/Martin.Raspaud/sentinel1", entry.title)
        if not os.path.exists(fname):
            url = "https://scihub.esa.int/dhus/odata/v1/Products('" + \
                entry.id + "')/$value"
            print "downloading", fname, "..."
            download_file(url, fname)
        else:
            print "skipping", fname



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2013, 2014 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Trollcast, server side.

Trollcasting is loosely based on the bittorrent concepts, and adapted to
satellite data.

Limitations:
 - HRPT specific at the moment

TODO:
 - Include files from a library, not only the currently written file to the
   list of scanlines
 - Implement choking
 - de-hardcode filename
"""
from __future__ import with_statement 

import logging
import os
from ConfigParser import ConfigParser, NoOptionError
from datetime import datetime, timedelta
from fnmatch import fnmatch
from glob import glob
from threading import Thread, Lock, Event
from urlparse import urlparse, urlunparse

import numpy as np
import time
from posttroll import strp_isoformat
from posttroll.message import Message
from posttroll.subscriber import Subscriber
from pyorbital.orbital import Orbital
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from zmq import Context, Poller, LINGER, PUB, REP, REQ, POLLIN, NOBLOCK

logger = logging.getLogger(__name__)

LINE_SIZE = 11090 * 2

CACHE_SIZE = 32

HRPT_SYNC = np.array([ 994, 1011, 437, 701, 644, 277, 452, 467, 833, 224, 694,
        990, 220, 409, 1010, 403, 654, 105, 62, 867, 75, 149, 320, 725, 668,
        581, 866, 109, 166, 941, 1022, 59, 989, 182, 461, 197, 751, 359, 704,
        66, 387, 238, 850, 746, 473, 573, 282, 6, 212, 169, 623, 761, 979, 338,
        249, 448, 331, 911, 853, 536, 323, 703, 712, 370, 30, 900, 527, 977,
        286, 158, 26, 796, 705, 100, 432, 515, 633, 77, 65, 489, 186, 101, 406,
        560, 148, 358, 742, 113, 878, 453, 501, 882, 525, 925, 377, 324, 589,
        594, 496, 972], dtype=np.uint16)
HRPT_SYNC_START = np.array([644, 367, 860, 413, 527, 149], dtype=np.uint16)

SATELLITES = {7: "NOAA 15",
              3: "NOAA 16",
              13: "NOAA 18",
              15: "NOAA 19"}

def timecode(tc_array):
    word = tc_array[0]
    day = word
    word = tc_array[1]
    msecs = ((127) & word) * 1024
    word = tc_array[2]
    msecs += word & 1023
    msecs *= 1024
    word = tc_array[3]
    msecs += word & 1023
    return timedelta(days=int(day/2 - 1), milliseconds=int(msecs))

class Holder(object):

    def __init__(self, configfile):
        self._holder = {}
        cfg = ConfigParser()
        cfg.read(configfile)
        host = cfg.get("local_reception", "localhost")
        hostname = cfg.get(host, "hostname")
        port = cfg.get(host, "pubport")
        self._addr = hostname + ":" + port

        self._station = cfg.get("local_reception", "station")

        self._context = Context()
        self._socket = self._context.socket(PUB)
        self._socket.bind("tcp://*:" + port)
        self._lock = Lock()
        self._send_lock = Lock()
        self._cache = []
        
        
    def __del__(self, *args, **kwargs):
        self._socket.close()

    def get(self, *args, **kwargs):
        return self._holder.get(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self._holder.__getitem__(*args, **kwargs)
    
    def send_have(self, satellite, utctime, elevation):
        """Sends 'have' message for *satellite*, *utctime*, *elevation*.
        """
        to_send = {}
        to_send["satellite"] = satellite
        to_send["timecode"] = utctime.isoformat()
        to_send["elevation"] = elevation
        to_send["origin"] = self._addr
        msg = Message('/oper/polar/direct_readout/' + self._station, "have",
                      to_send).encode()
        with self._send_lock:
            self._socket.send(msg)

    def send_heartbeat(self, next_pass_time="unknown"):
        to_send = {}
        to_send["next_pass_time"] = next_pass_time
        to_send["addr"] = self._addr
        msg =  Message('/oper/polar/direct_readout/' + self._station,
                       "heartbeat", to_send).encode()
        logger.debug("sending heartbeat: " + str(msg))
        with self._send_lock:
            self._socket.send(msg)

    def get_scanline(self, satellite, utctime):
        info = self._holder[satellite][utctime]
        if len(info) == 4:
            return info[3]
        else:
            url = urlparse(self._holder[satellite][utctime][1])
            with open(url.path, "rb") as fp_:
                fp_.seek(self._holder[satellite][utctime][0])
                return fp_.read(LINE_SIZE)


        
    def add_scanline(self, satellite, utctime, elevation, line_start, filename, line=None):
        """Adds the scanline to the server. Typically used by the client to
        signal newly received lines.
        """
        self._lock.acquire()
        try:
            if(satellite not in self._holder or
               utctime not in self._holder[satellite]):
                if line:
                    self._holder.setdefault(satellite,
                                            {})[utctime] = (line_start,
                                                            filename,
                                                            elevation,
                                                            line)
                    self._cache.append((satellite, utctime))
                    while len(self._cache) > CACHE_SIZE:
                        sat, deltime = self._cache[0]
                        del self._cache[0]
                        self._holder[sat][deltime] = \
                                                self._holder[sat][deltime][:3]
                        
                else:
                    self._holder.setdefault(satellite,
                                            {})[utctime] = (line_start,
                                                            filename,
                                                            elevation)
                self.send_have(satellite, utctime, elevation)
        finally:
            self._lock.release()
        

class FileStreamer(FileSystemEventHandler):
    """Get the updates from files.

    TODO: separate holder from file handling.
    """
    def __init__(self, holder, configfile, *args, **kwargs):
        FileSystemEventHandler.__init__(self, *args, **kwargs)
        self._file = None
        self._filename = ""
        self._where = 0
        self._satellite = ""
        self._orbital = None
        cfg = ConfigParser()
        cfg.read(configfile)
        self._coords = cfg.get("local_reception", "coordinates").split(" ")
        self._coords = [float(self._coords[0]),
                        float(self._coords[1]),
                        float(self._coords[2])]
        self._station = cfg.get("local_reception", "station")
        logger.debug("Station " + self._station +
                     " located at: " + str(self._coords))
        try:
            self._tle_files = cfg.get("local_reception", "tle_files")
        except NoOptionError:
            self._tle_files = None

        self._file_pattern = cfg.get("local_reception", "file_pattern")
        
        self.scanlines = holder

        self._warn = True
        
    def update_satellite(self, satellite):
        """Update satellite and renew the orbital instance.
        """
        if satellite != self._satellite:
            self._satellite = satellite
            if self._tle_files is not None:
                filelist = glob(self._tle_files)
                tle_file = max(filelist, key=lambda x: os.stat(x).st_mtime)
            else:
                tle_file = None
            self._orbital = Orbital(self._satellite.upper(), tle_file)

    def on_created(self, event):
        """Callback when file is created.
        """
        if event.src_path != self._filename:
            if self._filename:
                logger.info("Closing: " + self._filename)
            if self._file:
                self._file.close()
            self._file = None
            self._filename = ""
            self._where = 0
            self._satellite = ""
            self._warn = True
        logger.info("New file detected: " + event.src_path)


    def on_opened(self, event):
        """Callback when file is opened
        """
        fname = os.path.split(event.src_path)[1]
        if self._file is None and fnmatch(fname, self._file_pattern):
            logger.info("File opened: " + event.src_path)
            self._filename = event.src_path
            self._file = open(event.src_path, "rb")
            self._where = 0
            self._satellite = ""

            self._orbital = None
            self._warn = True

    def on_modified(self, event):
        self.on_opened(event)

        if event.src_path != self._filename:
            return
            
        self._file.seek(self._where)
        line = self._file.read(LINE_SIZE)
        while len(line) == LINE_SIZE:
            line_start = self._where
            self._where = self._file.tell()
            dtype = np.dtype([('frame_sync', '>u2', (6, )),
                              ('id', [('id', '>u2'),
                                      ('spare', '>u2')]),
                              ('timecode', '>u2', (4, )),
                              ('telemetry', [("ramp_calibration", '>u2', (5, )),
                                             ("PRT", '>u2', (3, )),
                                             ("ch3_patch_temp", '>u2'),
                                             ("spare", '>u2'),]),
                              ('back_scan', '>u2', (10, 3)),
                              ('space_data', '>u2', (10, 5)),
                              ('sync', '>u2'),
                              ('TIP_data', '>u2', (520, )),
                              ('spare', '>u2', (127, )),
                              ('image_data', '>u2', (2048, 5)),
                              ('aux_sync', '>u2', (100, ))])


            array = np.fromstring(line, dtype=dtype)
            if np.all(abs(HRPT_SYNC_START - 
                          array["frame_sync"]) > 1):
                array = array.newbyteorder()
                
            # FIXME: this means server can only share 1 year of hrpt data.
            now = datetime.utcnow()
            year = now.year
            utctime = datetime(year, 1, 1) + timecode(array["timecode"][0])
            if utctime > now:
                # Can't have data from the future... yet :)
                utctime = (datetime(year - 1, 1, 1)
                           + timecode(array["timecode"][0]))

            # Check that we receive real-time data
            if not (np.all(array['aux_sync'] == HRPT_SYNC) and
                    np.all(array['frame_sync'] == HRPT_SYNC_START)):
                logger.info("Garbage line: " + str(utctime))
                line = self._file.read(LINE_SIZE)
                continue

            if (now - utctime).days > 7 and self._warn:
                logger.warning("Data is more than a week old: " + str(utctime))
                self._warn = False
                
            satellite = SATELLITES[((array["id"]["id"] >> 3) & 15)[0]]
            self.update_satellite(satellite)

            elevation = self._orbital.get_observer_look(utctime,
                                                        *self._coords)[1]
            logger.debug("Got line " + utctime.isoformat() + " "
                         + self._satellite + " "
                         + str(elevation))


            
            # TODO:
            # - serve also already present files
            # - timeout and close the file
            self.scanlines.add_scanline(self._satellite, utctime,
                                        elevation, line_start, self._filename,
                                        line)

            line = self._file.read(LINE_SIZE)

        self._file.seek(self._where)        

class MirrorStreamer(Thread):
    """Act as a relay...
    """

    def __init__(self, holder, configfile):
        Thread.__init__(self)
        
        self.scanlines = holder
        
        cfg = ConfigParser()
        cfg.read(configfile)
        host = cfg.get("local_reception", "mirror")
        hostname = cfg.get(host, "hostname")
        port = cfg.get(host, "pubport")
        rport = cfg.get(host, "reqport")
        address = "tcp://" + hostname + ":" + port
        self._sub = Subscriber([address], "hrpt 0")
        self._reqaddr = "tcp://" + hostname + ":" + rport

    def run(self):

        for message in self._sub.recv(1):
            if message is None:
                continue
            if(message.type == "have"):
                logger.debug("Relaying " + str(message.data["timecode"]))
                self.scanlines.add_scanline(message.data["satellite"],
                                            strp_isoformat(message.data["timecode"]),
                                            message.data["elevation"],
                                            None,
                                            self._reqaddr)
            # Don't forward the heartbeat!
            #if(message.type) == "heartbeat":
            #    self.scanlines._socket.send(str(message))
                
    def stop(self):
        """Stop streaming.
        """
        self._sub.stop()
        
class Looper(object):

    def __init__(self):
        self._loop = True

    def stop(self):
        self._loop = False

class Socket(object):
    def __init__(self, addr, stype):
        self._context = Context()
        self._socket = self._context.socket(stype)
        if stype in [REP, PUB]:
            self._socket.bind(addr)
        else:
            self._socket.connect(addr)

    def __del__(self, *args, **kwargs):
        self._socket.close()

class SocketLooper(Socket, Looper):
    
    def __init__(self, *args, **kwargs):
        Looper.__init__(self)
        Socket.__init__(self, *args, **kwargs)

class SocketLooperThread(SocketLooper, Thread):
    def __init__(self, *args, **kwargs):
        Thread.__init__(self)
        SocketLooper.__init__(self, *args, **kwargs)

class Heart(Thread):
    def __init__(self, holder, *args, **kwargs):
        Thread.__init__(self, *args, **kwargs)
        self.loop = True
        self.holder = holder
        self.event = Event()
        self.interval = 300

    def run(self):
        while self.loop:
            self.holder.send_heartbeat()
            self.event.wait(self.interval)

    def stop(self):
        self.loop = False
        self.event.set()

class Responder(SocketLooperThread):

    # TODO: this should not respond to everyone. It should check if the
    # requester is listed in the configuration file...
    
    def __init__(self, holder, configfile, *args, **kwargs):
        SocketLooperThread.__init__(self, *args, **kwargs)
        self._holder = holder
        self._loop = True

        cfg = ConfigParser()
        cfg.read(configfile)
        self._station = cfg.get("local_reception", "station")

        self.mirrors = {}


    def __del__(self, *args, **kwargs):
        self._socket.close()
        for mirror in self.mirrors.values():
            mirror.close()
            
        
    def forward_request(self, address, message):
        """Forward a request to another server.
        """
        # FIXME: shouldn't we have an open socket at all times ?
        if address not in self.mirrors:
            context = Context()
            socket = context.socket(REQ)
            socket.connect(address)
            self.mirrors[address] = socket
        else:
            socket = self.mirrors[address]
        socket.send(str(message))
        return socket.recv()


    def run(self):
        poller = Poller()
        poller.register(self._socket, POLLIN)
        
        while self._loop:
            socks = dict(poller.poll(timeout=2000))
            if self._socket in socks and socks[self._socket] == POLLIN:
                message = Message(rawstr=self._socket.recv(NOBLOCK))
                logger.debug("Request: " + str(message))
                # send list of scanlines
                if(message.type == "request" and
                   message.data["type"] == "scanlines"):
                    sat = message.data["satellite"]
                    epoch = "1950-01-01T00:00:00"
                    start_time = strp_isoformat(message.data.get("start_time",
                                                                 epoch))
                    end_time = strp_isoformat(message.data.get("end_time",
                                                               epoch))

                    resp = Message('/oper/polar/direct_readout/' + self._station,
                                   "scanlines",
                                   [(utctime.isoformat(),
                                     self._holder[sat][utctime][2])
                                    for utctime in self._holder.get(sat, [])
                                    if utctime >= start_time
                                    and utctime <= end_time])
                    self._socket.send(str(resp))

                # send one scanline
                elif(message.type == "request" and
                     message.data["type"] == "scanline"):
                    sat = message.data["satellite"]
                    utctime = strp_isoformat(message.data["utctime"])
                    try:
                        url = urlparse(self._holder[sat][utctime][1])
                    except KeyError:
                        resp = Message('/oper/polar/direct_readout/'
                                       + self._station,
                                       "missing")
                    else:
                        if url.scheme in ["", "file"]: # data is locally stored.
                            resp = Message('/oper/polar/direct_readout/'
                                           + self._station,
                                           "scanline",
                                           self._holder.get_scanline(sat, utctime),
                                           binary=True)
                        else: # it's the address of a remote server.
                            resp = self.forward_request(urlunparse(url),
                                                        message)
                    self._socket.send(str(resp))

                # take in a new scanline
                elif(message.type == "notice" and
                     message.data["type"] == "scanline"):
                    sat = message.data["satellite"]
                    utctime = message.data["utctime"]
                    elevation = message.data["elevation"]
                    filename = message.data["filename"]
                    line_start = message.data["file_position"]
                    utctime = strp_isoformat(utctime)
                    self._holder.add_scanline(sat, utctime, elevation,
                                              line_start, filename)
                    resp = Message('/oper/polar/direct_readout/'
                                       + self._station,
                                       "notice",
                                       "ack")
                    self._socket.send(str(resp))
                elif(message.type == "ping"):
                    resp = Message('/oper/polar/direct_readout/'
                                   + self._station,
                                   "pong",
                                   {"station": self._station})
                    self._socket.send(str(resp))

                if resp.binary:
                    logger.debug("Response: " + " ".join(str(resp).split()[:6]))
                else:
                    logger.debug("Response: " + str(resp))
                
    def stop(self):
        self._loop = False


def serve(configfile):
    """Serve forever.
    """

    scanlines = Holder(configfile)
    heartbeat = Heart(scanlines)
    fstreamer = FileStreamer(scanlines, configfile)
    notifier = Observer()
    cfg = ConfigParser()
    cfg.read(configfile)
    path = cfg.get("local_reception", "data_dir")
    notifier.schedule(fstreamer, path, recursive=False)
    

    local_station = cfg.get("local_reception", "localhost")
    responder_port = cfg.get(local_station, "reqport")
    responder = Responder(scanlines, configfile,
                          "tcp://*:" + responder_port, REP)
    responder.start()
    heartbeat.start()
    mirror = None
    try:
        mirror = MirrorStreamer(scanlines, configfile)
        mirror.start()
    except NoOptionError:
        pass
    
    notifier.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        notifier.stop()
    
    responder.stop()
    heartbeat.stop()
    notifier.join()

    if mirror is not None:
        mirror.stop()
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read schedule files.
"""

from datetime import datetime, timedelta

formats = ["scisys"]
satellites = ["NOAA 15", "NOAA 16", "NOAA 18", "NOAA 19"]


def scisys(filename):
    """Read a scisys schedule
    """
    with open(filename) as fp_:
        while True:
            line = fp_.readline()
            try:
                if line[0] in ["\n", " ", "!", "#"]:
                    continue
            except IndexError:
                break
            elts = line[16:].split()
            sat = line[:16].strip()
            rec = elts[8] == "Y"
            if rec and sat in satellites:
                rise, fall = "".join(elts[1:3]), "".join(elts[3:5])
                rise = datetime.strptime(rise, "%Y%m%d%H%M%S")
                fall = datetime.strptime(fall, "%Y%m%d%H%M%S")
                yield (rise, sat, fall)


def kongsberg_metno(filename):
    """Read a kongsberg schedule
    """
    with open(filename) as fp_:
        while True:
            line = fp_.readline()
            if line.startswith(" No."):
                continue

            elts = line.split()
            try:
                sat = elts[2].replace("_", " ")
            except IndexError:
                break
            rec = elts[13].startswith("Schedule")
            if rec and sat in satellites:
                rise = " ".join((elts[1], elts[5]))
                rise = datetime.strptime(rise, "%Y%m%d %H:%M:%S")
                mins, secs = elts[8].split(":")
                fall = rise + timedelta(minutes=int(mins), seconds=int(secs))
                yield (rise, sat, fall)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""New version of the trollcast server

TODO:
 - add lines when local client gets data (if missing)
 - check that mirror server is alive
"""

from ConfigParser import ConfigParser, NoOptionError
from zmq import Context, Poller, LINGER, PUB, REP, POLLIN, NOBLOCK, SUB, SUBSCRIBE, ZMQError
from threading import Thread, Event, Lock, Timer
from posttroll.message import Message
from posttroll import context
from pyorbital.orbital import Orbital
import logging
import time
from datetime import datetime, timedelta
from posttroll import strp_isoformat
from fnmatch import fnmatch
from pyinotify import (WatchManager, ProcessEvent, ThreadedNotifier,
                       IN_MODIFY, IN_OPEN, IN_CLOSE_WRITE)
from urlparse import urlparse
import os
import numpy as np
from glob import glob
from trollcast import schedules
from trollsift.parser import globify, parse

logger = logging.getLogger(__name__)

subject = None
coords = None
tle_files = None


def get_f_elev(satellite):
    """Get the elevation function for a given satellite
    """

    if tle_files is not None:
        filelist = glob(tle_files)
        tle_file = max(filelist, key=lambda x: os.stat(x).st_mtime)
    else:
        tle_file = None

    orb = Orbital(satellite.upper(), tle_file)

    def f_elev(utctime):
        """Get the elevation for the given *utctime*.
        """
        return orb.get_observer_look(utctime, *coords)[1]
    f_elev.satellite = satellite
    return f_elev


class CADU(object):

    """The cadu reader class
    """
    @staticmethod
    def is_it(data):
        return False


class HRPT(object):

    """The hrpt reader class
    """
    dtype = np.dtype([('frame_sync', '>u2', (6, )),
                      ('id', [('id', '>u2'),
                              ('spare', '>u2')]),
                      ('timecode', '>u2', (4, )),
                      ('telemetry', [("ramp_calibration", '>u2', (5, )),
                                     ("PRT", '>u2', (3, )),
                                     ("ch3_patch_temp", '>u2'),
                                     ("spare", '>u2'), ]),
                      ('back_scan', '>u2', (10, 3)),
                      ('space_data', '>u2', (10, 5)),
                      ('sync', '>u2'),
                      ('TIP_data', '>u2', (520, )),
                      ('spare', '>u2', (127, )),
                      ('image_data', '>u2', (2048, 5)),
                      ('aux_sync', '>u2', (100, ))])

    hrpt_sync = np.array([994, 1011, 437, 701, 644, 277, 452, 467, 833, 224,
                          694, 990, 220, 409, 1010, 403, 654, 105, 62, 867,
                          75, 149, 320, 725, 668, 581, 866, 109, 166, 941,
                          1022, 59, 989, 182, 461, 197, 751, 359, 704, 66,
                          387, 238, 850, 746, 473, 573, 282, 6, 212, 169, 623,
                          761, 979, 338, 249, 448, 331, 911, 853, 536, 323,
                          703, 712, 370, 30, 900, 527, 977, 286, 158, 26, 796,
                          705, 100, 432, 515, 633, 77, 65, 489, 186, 101, 406,
                          560, 148, 358, 742, 113, 878, 453, 501, 882, 525,
                          925, 377, 324, 589, 594, 496, 972], dtype=np.uint16)

    hrpt_sync_start = np.array([644, 367, 860, 413, 527, 149], dtype=np.uint16)

    satellites = {7: "NOAA 15",
                  3: "NOAA 16",
                  13: "NOAA 18",
                  15: "NOAA 19"}

    line_size = 11090 * 2

    @staticmethod
    def is_it(data):
        return True

    @staticmethod
    def timecode(tc_array):
        """HRPT timecode reading
        """
        word = tc_array[0]
        day = word
        word = tc_array[1]
        msecs = ((127) & word) * 1024
        word = tc_array[2]
        msecs += word & 1023
        msecs *= 1024
        word = tc_array[3]
        msecs += word & 1023
        return timedelta(days=int(day / 2 - 1), milliseconds=int(msecs))

    def __init__(self, sat, reftime):
        self.sat = sat
        self.reftime = reftime
        self.count = 0
        self.time_threshold = timedelta(seconds=600)
        self.reftimes = []
        self.last_time = None
        self.lags = [0]

    def read(self, data, f_elev=None):
        """Read hrpt data.
        """

        now = datetime.utcnow()
        year = now.year

        for i, line in enumerate(np.fromstring(data, dtype=self.dtype,
                                               count=len(data) / self.line_size)):
            days = self.timecode(line["timecode"])
            utctime = datetime(year, 1, 1) + days

            if utctime > now:
                # Can't have data from the future... yet :)
                utctime = datetime(year - 1, 1, 1) + days

            qual = (np.sum(line['aux_sync'] == self.hrpt_sync) +
                    np.sum(line['frame_sync'] == self.hrpt_sync_start))
            qual = 100 * (qual / 106.0)

            qual -= np.sum(abs(np.diff(line['image_data']
                                       [:, 4].astype(np.int16))) > 200)

            if np.sum(line['frame_sync'] == self.hrpt_sync_start) < 5:
                logger.info("Frame sync not in place, setting quality to 0")
                qual = 0

            # Take care of noisy time codes.

            new_time = datetime.now()

            # Check if we missed some scanlines
            if(self.last_time is not None):
                time_diff = (new_time - self.last_time)
                seconds = (time_diff.days * 24 * 60 * 60
                           + time_diff.seconds
                           + time_diff.microseconds / 1000000.0)
                self.lags.append(seconds)
                if(abs(utctime - (self.reftime +
                                  timedelta(seconds=self.count / 6.0 +
                                            max(self.lags))))
                   < timedelta(seconds=15)):
                    logger.debug(
                        "Looks like we lost some scanlines, adjusting %s counts.", seconds * 6.0)

                    self.count += int(max(self.lags) * 6.0)
                    self.time_threshold = max(timedelta(seconds=15),
                                              self.time_threshold)
                    self.reftimes = []
                    self.lags = [0]

            # Adjust reference time upon good quality scanlines
            if qual >= 99.9:
                # adjusting reftime:
                self.reftimes.append(
                    utctime - timedelta(seconds=self.count / 6.0))
                if (len(self.reftimes) > 50 and
                        self.time_threshold > timedelta(milliseconds=50)):
                    old_ref = self.reftime

                    hist, boundaries = np.histogram([time.mktime(reftime.timetuple())
                                                     for reftime in self.reftimes],
                                                    100)
                    idx = np.argmax(hist)
                    self.reftime = datetime.fromtimestamp(
                        (boundaries[idx] + boundaries[idx + 1]) / 2)
                    self.time_threshold = timedelta(milliseconds=50)
                    logger.info("Setting up new reftime %s (was %s)",
                                str(self.reftime),
                                str(old_ref))
            elif (abs(utctime - (self.reftime +
                                 timedelta(seconds=self.count / 6.0)))
                    > self.time_threshold):
                # Check if scanline is within expected range
                logger.debug("Spurious time for scanline %s (should be %s)",
                             str(utctime),
                             str((self.reftime + timedelta(seconds=self.count / 6.0))))
                qual = 0

            self.count += 1
            self.last_time = new_time

            qual = max(0, qual)

            logger.info("Quality " + str(qual))

            if qual != 100:
                logger.info("Degraded line: " + str(utctime))
                if f_elev is None:
                    satellite = "unknown"
                    yield ((satellite, utctime, None, qual,
                            data[self.line_size * i: self.line_size * (i + 1)]),
                           self.line_size * (i + 1), f_elev)
                    continue
                else:
                    satellite = f_elev.satellite
            else:
                try:
                    satellite = self.satellites[((line["id"]["id"] >> 3) & 15)]
                except KeyError:
                    satellite = "unknown"

            if f_elev is None:
                if satellite != "unknown":
                    f_elev = get_f_elev(satellite)
                    elevation = f_elev(utctime)
                else:
                    elevation = -180
            else:
                elevation = f_elev(utctime)

            logger.debug("Got line " + utctime.isoformat() + " "
                         + satellite + " "
                         + str(elevation))

            # TODO:
            # - serve also already present files
            # - timeout and close the file

            yield ((satellite, utctime, elevation, qual,
                    data[self.line_size * i: self.line_size * (i + 1)]),
                   self.line_size * (i + 1), f_elev)


FORMATS = [CADU, HRPT]


class FileWatcher(Thread):

    def __init__(self, holder, uri, schedule_reader):
        Thread.__init__(self)
        self._loop = True
        self._error_event = Event()
        self._wm = WatchManager()
        self._holder = holder
        self._uri = uri
        self._schedule_reader = schedule_reader
        self._notifier = ThreadedNotifier(self._wm,
                                          _EventHandler(self._holder,
                                                        self._uri,
                                                        self._schedule_reader,
                                                        self._error_event))
        self._path, self._pattern = os.path.split(urlparse(uri).path)

    def start(self):
        """Start the file watcher
        """
        self._notifier.start()
        self._wm.add_watch(self._path, IN_OPEN | IN_CLOSE_WRITE | IN_MODIFY)
        Thread.start(self)

    def run(self):
        while self._loop:
            if self._error_event.wait(1):
                self._error_event.clear()
                self._notifier.stop()
                del self._notifier
                self._notifier = ThreadedNotifier(
                    self._wm,
                    _EventHandler(self._holder,
                                  self._uri,
                                  self._schedule_reader,
                                  self._error_event))
                self._notifier.start()
                self._wm.add_watch(
                    self._path, IN_OPEN | IN_CLOSE_WRITE | IN_MODIFY)

    def stop(self):
        """Stop the file watcher
        """
        self._notifier.stop()
        self._loop = False


class _EventHandler(ProcessEvent):

    """Watch files
    """

    def __init__(self, holder, uri, schedule_reader, error_event):
        ProcessEvent.__init__(self)
        self._holder = holder
        self._uri = uri
        self._current_pass = None
        self._schedule_reader = schedule_reader
        self._loop = True
        self._error_event = error_event

        self._path, self._pattern = os.path.split(urlparse(self._uri).path)

        self._readers = {}
        self._fp = None
        self._receiving = False
        self._timer = None
        self.sat = None
        self.time = None
        self.current_event = None
        self._pass_end_timer = None

        if self._schedule_reader.next_pass:
            next_pass_in = (self._schedule_reader.next_pass[0]
                            - datetime.utcnow())
            if next_pass_in.seconds > 0:
                self._timer = Timer(next_pass_in.seconds + 5,
                                    self.error)
                self._timer.start()

    def error(self):
        logger.critical("Reception expected but not started")
        self._error_event.set()

    def set_reception_active(self, event):
        self._receiving = True
        if self._timer is not None:
            self._timer.cancel()
        if self._pass_end_timer is not None:
            self._pass_end_timer.cancel()
        self._pass_end_timer = Timer(60, self.clean_up, event)

    def stop_receiving(self):
        self._receiving = False
        if self._schedule_reader.next_pass:
            next_pass_in = (self._schedule_reader.next_pass[0]
                            - datetime.utcnow())
            self._timer = Timer(next_pass_in.seconds + 5,
                                self.error)
            self._timer.start()

    def _reader(self, pathname, current_pass):
        """Read the file
        """
        try:
            try:
                filereader, position, f_elev = self._readers[pathname]
            except KeyError:
                position = 0
                f_elev = None

            self._fp.seek(position)

            data = self._fp.read()

            if position == 0:
                for filetype in FORMATS:
                    if filetype.is_it(data):
                        filereader = filetype(self.sat, self.time)

            for elt, offset, f_elev in filereader.read(data, f_elev):
                self._readers[pathname] = filereader, position + offset, f_elev
                if self.sat is not None:
                    elt = list(elt)
                    if elt[0] != self.sat:
                        logger.debug("Satellite id scrambled, "
                                     "lowering quality score.")
                        elt[3] -= 1
                    elt[0] = self.sat
                elif current_pass is not None:
                    elt = list(elt)
                    if elt[1] > current_pass[2] or elt[1] < current_pass[0]:
                        logger.debug("line %s doesn't match current pass %s",
                                     str(elt[:2]), str(current_pass))
                        continue
                    if elt[0] != current_pass[1]:
                        logger.debug("Satellite id scrambled, "
                                     "lowering quality score.")
                        elt[3] -= 1
                    elt[0] = current_pass[1]
                yield elt
        except IOError, err:
            logger.warning("Can't read file: " + str(err))
            return

    def process_IN_OPEN(self, event):
        """When the file opens.
        """

        fname = os.path.basename(event.pathname)

        if not fnmatch(fname, globify(self._pattern)):
            logger.debug("Ignoring %s", event.pathname)
            return False

        if self.current_event is None:
            self.current_event = event
        elif(event.pathname != self.current_event.pathname):
            self.clean_up(self.current_event)
            self.current_event = event

        if self._fp is None:
            self._fp = open(event.pathname)
            self._current_pass = self._schedule_reader.next_pass
            info = parse(self._pattern, fname)
            try:
                self.sat = " ".join((info["platform"], info["number"]))
                self.time = info["utctime"]
            except KeyError:
                logger.info("Could not retrieve satellite name from filename")

        self.set_reception_active(event)
        return self._fp is not None

    def process_IN_MODIFY(self, event):
        """File has been modified, read it !
        """

        if not self.process_IN_OPEN(event):
            return

        logger.debug("File modified! %s", event.pathname)

        fname = os.path.basename(event.pathname)

        if not fnmatch(fname, globify(self._pattern)):
            return

        self.set_reception_active(event)

        for sat, key, elevation, qual, data in self._reader(event.pathname,
                                                            self._current_pass):
            if qual > 0:
                self._holder.add(sat, key, elevation, qual, data)

    # def process_IN_CLOSE_WRITE(self, event):
    def clean_up(self, event):
        """Clean up.
        """
        fname = os.path.basename(event.pathname)

        if not fnmatch(fname, globify(self._pattern)):
            return

        if self._fp is not None:
            self._fp.close()
        else:
            logger.warning("File descriptor is None for %s", event.pathname)
        self._fp = None
        self._schedule_reader.get_next_pass()
        self.stop_receiving()
        try:
            del self._readers[event.pathname]
        except KeyError:
            logger.info("No reader defined for %s", str(event.pathname))

from trollcast.client import SimpleRequester, REQ_TIMEOUT


class _MirrorGetter(object):

    """Gets data from the mirror when needed.
    """

    def __init__(self, req, sat, key):
        self._req = req
        self._sat = sat
        self._key = key
        self._data = None

    def get_data(self):
        """Get the actual data from the server we're mirroring
        """
        if self._data is not None:
            return self._data

        logger.debug("Grabbing scanline from mirror")
        reqmsg = Message(subject,
                         'request',
                         {"type": "scanline",
                          "satellite": self._sat,
                          "utctime": self._key})
        rep = self._req.send_and_recv(str(reqmsg), REQ_TIMEOUT)

        if rep and rep.data:
            self._data = rep.data
            logger.debug("Retrieved scanline from mirror successfully")
        else:
            self._data = None
            logger.warning("Empty scanline from mirror!")
        return self._data

    def __str__(self):
        return str(self.get_data())

    def __add__(self, other):
        return str(self) + other

    def __radd__(self, other):
        return other + str(self)


class MirrorWatcher(Thread):

    """Watches a other server.
    """

    def __init__(self, holder, host, pubport, reqport, sched):
        Thread.__init__(self)
        self._holder = holder
        self._pubaddress = "tcp://" + host + ":" + str(pubport)
        self._reqaddress = "tcp://" + host + ":" + str(reqport)

        self._req = SimpleRequester(host, reqport)

        self._subsocket = context.socket(SUB)
        self._subsocket.connect(self._pubaddress)
        self._subsocket.setsockopt(SUBSCRIBE, "pytroll")
        self._poller = Poller()
        self._poller.register(self._subsocket, POLLIN)
        self._lock = Lock()
        self._loop = True
        self._sched = sched

    def run(self):
        last_hb = datetime.now()
        minutes = 2
        while self._loop:
            if datetime.now() - last_hb > timedelta(minutes=minutes):
                logger.error("No heartbeat from " + str(self._pubaddress))
                last_hb = datetime.now()
                minutes = 1440
            socks = dict(self._poller.poll(2000))
            if (socks and
                    self._subsocket in socks and
                    socks[self._subsocket] == POLLIN):
                message = Message.decode(self._subsocket.recv())
            else:
                continue
            if message.type == "have":
                sat = message.data["satellite"]
                key = strp_isoformat(message.data["timecode"])
                elevation = message.data["elevation"]
                quality = message.data.get("quality", 100)
                data = _MirrorGetter(self._req, sat, key)
                self._holder.add(sat, key, elevation, quality, data)
            if message.type == "heartbeat":
                logger.debug("Got heartbeat from " + str(self._pubaddress)
                             + ": " + str(message))
                self._sched.mirror_next_pass = message.data["next_pass"]
                last_hb = datetime.now()
                minutes = 2

    def stop(self):
        """Stop the watcher
        """
        self._loop = False
        self._req.stop()
        self._subsocket.setsockopt(LINGER, 0)
        self._subsocket.close()


class DummyWatcher(Thread):

    """Dummy watcher for test purposes
    """

    def __init__(self, holder, uri):
        Thread.__init__(self)
        self._holder = holder
        self._uri = uri
        self._loop = True
        self._event = Event()

    def run(self):
        while self._loop:
            self._holder.add("NOAA 17", datetime.utcnow(),
                             18, 100, "dummy data")
            self._event.wait(self._uri)

    def stop(self):
        """Stop adding stuff
        """
        self._loop = False
        self._event.set()


class Cleaner(Thread):

    """Dummy watcher for test purposes
    """

    def __init__(self, holder, delay):
        Thread.__init__(self)
        self._holder = holder
        self._interval = 60
        self._delay = delay
        self._loop = True
        self._event = Event()

    def clean(self):
        """Clean the db
        """
        logger.debug("Cleaning")
        for sat in self._holder.sats():
            satlines = self._holder.get_sat(sat)
            for key in sorted(satlines):
                if key < datetime.utcnow() - timedelta(hours=self._delay):
                    self._holder.delete(sat, key)

    def run(self):
        while self._loop:
            self.clean()
            self._event.wait(self._interval)

    def stop(self):
        """Stop adding stuff
        """
        self._loop = False
        self._event.set()


class Holder(object):

    """The mighty data holder
    """

    def __init__(self, pub, origin):
        self._data = {}
        self._pub = pub
        self._origin = origin
        self._lock = Lock()

    def delete(self, sat, key):
        """Delete item
        """
        logger.debug("Removing from memory: " + str((sat, key)))
        with self._lock:
            del self._data[sat][key]

    def get_sat(self, sat):
        """Get the data for a given satellite *sat*.
        """
        return self._data[sat]

    def sats(self):
        """return the satellites in store.
        """
        return self._data.keys()

    def get(self, sat, key):
        """get the value of *sat* and *key*
        """
        with self._lock:
            return self._data[sat][key]

    def get_data(self, sat, key):
        """get the data of *sat* and *key*
        """
        return self.get(sat, key)[2]

    def add(self, sat, key, elevation, qual, data):
        """Add some data.
        """
        with self._lock:
            self._data.setdefault(sat, {})[key] = elevation, qual, data
        logger.debug("Got stuff for " + str((sat, key, elevation, qual)))
        self.have(sat, key, elevation, qual)

    def have(self, sat, key, elevation, qual):
        """Tell the world about our new data.
        """
        to_send = {}
        to_send["satellite"] = sat
        to_send["timecode"] = key
        to_send["elevation"] = elevation
        to_send["quality"] = qual
        to_send["origin"] = self._origin
        logger.info(str(to_send))

        msg = Message(subject, "have", to_send).encode()
        self._pub.send(msg)


class Publisher(object):

    """Publish stuff.
    """

    def __init__(self, port):
        self._socket = context.socket(PUB)
        self._socket.bind("tcp://*:" + str(port))
        self._lock = Lock()

    def send(self, message):
        """Publish something
        """
        with self._lock:
            self._socket.send(str(message))

    def stop(self):
        """Stop publishing.
        """
        with self._lock:
            self._socket.setsockopt(LINGER, 0)
            self._socket.close()


class ScheduleReader(object):

    """Reads and handles a schedule
    """

    def __init__(self, filename, fileformat):
        """

        Arguments:
        - `filename`: schedule filename
        - `fileformat`: schedule format
        """
        self._filename = filename
        self._fileformat = fileformat
        self.next_pass = None
        self.mirror_next_pass = None

    def get_next_pass(self):
        """Get the next pass from the schedule
        """
        try:
            fun = getattr(schedules, self._fileformat)
        except TypeError:
            return None
        logger.debug("using %s", str(fun))
        now = datetime.utcnow()
        for overpass in fun(self._filename):
            if overpass[2] > now:
                self.next_pass = overpass
                return overpass


class Heart(Thread):

    """Send heartbeats once in a while.
    """

    def __init__(self, pub, address, interval, schedule_reader):
        Thread.__init__(self)
        self._loop = True
        self._event = Event()
        self._address = address
        self._pub = pub
        self._schedule_reader = schedule_reader
        self._interval = interval

    def run(self):
        while self._loop:
            to_send = {}
            keys = ("start_time", "platform_name", "end_time")
            next_pass = self._schedule_reader.get_next_pass()
            mirror_next_pass = self._schedule_reader.mirror_next_pass
            passes = []
            if next_pass is not None:
                passes.append(dict(zip(keys, next_pass)))
            if mirror_next_pass is not None:
                passes.extend(mirror_next_pass)
            if passes:
                to_send["next_pass"] = sorted(passes, key=(lambda x: x["start_time"]))
            to_send["addr"] = self._address
            msg = Message(subject, "heartbeat", to_send).encode()
            logger.debug("sending heartbeat: " + str(msg))
            self._pub.send(msg)
            self._event.wait(self._interval)

    def stop(self):
        """Cardiac arrest
        """
        self._loop = False
        self._event.set()


class RequestManager(Thread):

    """Manage requests.
    """

    def __init__(self, holder, port, station):
        Thread.__init__(self)

        self._holder = holder
        self._loop = True
        self._port = port
        self._station = station
        self._lock = Lock()
        self._socket = context.socket(REP)
        self._socket.bind("tcp://*:" + str(self._port))
        self._poller = Poller()
        self._poller.register(self._socket, POLLIN)

    def send(self, message):
        """Send a message
        """
        if message.binary:
            logger.debug("Response: " + " ".join(str(message).split()[:6]))
        else:
            logger.debug("Response: " + str(message))
        self._socket.send(str(message))

    def pong(self):
        """Reply to ping
        """
        return Message(subject, "pong", {"station": self._station})

    def scanline(self, message):
        """Reply to scanline request
        """
        sat = message.data["satellite"]
        key = strp_isoformat(message.data["utctime"])
        try:
            data = self._holder.get_data(sat, key)
        except KeyError:
            resp = Message(subject, "missing")
        else:
            resp = Message(subject, "scanline", data, binary=True)
        return resp

    def notice(self, message):
        """Reply to notice message
        """
        del message
        return Message(subject, "ack")

    def unknown(self, message):
        """Reply to any unknown request.
        """
        del message
        return Message(subject, "unknown")

    def run(self):
        while self._loop:
            try:
                socks = dict(self._poller.poll(timeout=2000))
            except ZMQError:
                logger.info("Poller interrupted.")
                continue
            if self._socket in socks and socks[self._socket] == POLLIN:
                logger.debug("Received a request, waiting for the lock")
                with self._lock:
                    message = Message(rawstr=self._socket.recv(NOBLOCK))
                    logger.debug("processing request: " + str(message))
                    reply = Message(subject, "error")
                    try:
                        if message.type == "ping":
                            reply = self.pong()
                        elif (message.type == "request" and
                              message.data["type"] == "scanline"):
                            reply = self.scanline(message)
                        elif (message.type == "notice" and
                              message.data["type"] == "scanline"):
                            reply = self.notice(message)
                        else:  # unknown request
                            reply = self.unknown(message)
                    except:
                        logger.exception("Something went wrong"
                                         " when processing the request:")
                    finally:
                        self.send(reply)
                logger.debug("Lock released from manager")
            else:  # timeout
                pass

    def stop(self):
        """Stop the request manager.
        """
        self._loop = False
        self._socket.setsockopt(LINGER, 0)
        self._socket.close()


def set_subject(station):
    global subject
    subject = '/oper/polar/direct_readout/' + station


def serve(configfile):
    """Serve forever.
    """

    try:
        cfg = ConfigParser()
        cfg.read(configfile)

        host = cfg.get("local_reception", "localhost")

        # for messages
        station = cfg.get("local_reception", "station")
        set_subject(station)

        # for elevation
        global coords
        coords = cfg.get("local_reception", "coordinates")
        coords = [float(coord) for coord in coords.split()]

        global tle_files

        try:
            tle_files = cfg.get("local_reception", "tle_files")
        except NoOptionError:
            tle_files = None

        # publisher
        pubport = cfg.getint(host, "pubport")
        pub = Publisher(pubport)

        # schedule reader
        try:
            sched = ScheduleReader(cfg.get("local_reception", "schedule_file"),
                                   cfg.get("local_reception", "schedule_format"))
            sched.get_next_pass()
        except NoOptionError:
            logger.warning("No schedule file given")
            sched = ScheduleReader(None, None)

        # heart
        hostname = cfg.get(host, "hostname")
        pubaddress = hostname + ":" + str(pubport)
        heart = Heart(pub, pubaddress, 30, sched)
        heart.start()

        # holder
        holder = Holder(pub, pubaddress)

        # cleaner

        cleaner = Cleaner(holder, 1)
        cleaner.start()

        # watcher
        #watcher = DummyWatcher(holder, 2)
        path = cfg.get("local_reception", "data_dir")
        watcher = None

        if not os.path.exists(path):
            logger.warning(
                path + " doesn't exist, not getting data from files")
        else:
            pattern = cfg.get("local_reception", "file_pattern", raw=True)
            watcher = FileWatcher(holder, os.path.join(path, pattern), sched)
            watcher.start()

        mirror_watcher = None
        try:
            mirror = cfg.get("local_reception", "mirror")
        except NoOptionError:
            pass
        else:
            pubport_m = cfg.getint(mirror, "pubport")
            reqport_m = cfg.getint(mirror, "reqport")
            host_m = cfg.get(mirror, "hostname")
            mirror_watcher = MirrorWatcher(holder,
                                           host_m, pubport_m, reqport_m,
                                           sched)
            mirror_watcher.start()

        # request manager
        reqport = cfg.getint(host, "reqport")
        reqman = RequestManager(holder, reqport, station)
        reqman.start()

        while True:
            time.sleep(10000)

    except KeyboardInterrupt:
        pass
    except:
        logger.exception("There was an error!")
        raise
    finally:
        try:
            reqman.stop()
        except UnboundLocalError:
            pass

        try:
            if mirror_watcher is not None:
                mirror_watcher.stop()
        except UnboundLocalError:
            pass

        try:
            if watcher is not None:
                watcher.stop()
        except UnboundLocalError:
            pass
        try:
            cleaner.stop()
        except UnboundLocalError:
            pass
        try:
            heart.stop()
        except UnboundLocalError:
            pass
        try:
            pub.stop()
        except UnboundLocalError:
            pass
        try:
            context.term()
        except ZMQError:
            pass


if __name__ == '__main__':

    import sys

    ch1 = logging.StreamHandler()
    ch1.setLevel(logging.DEBUG)

    formatter = logging.Formatter('[%(levelname)s %(name)s %(asctime)s] '
                                  '%(message)s')
    ch1.setFormatter(formatter)

    logging.getLogger('').setLevel(logging.DEBUG)
    logging.getLogger('').addHandler(ch1)
    logger = logging.getLogger("trollcast_server")

    # Test hrpt reading

    coords = [16.148649, 58.581844, 0.052765]

    reftime = datetime.strptime(
        os.path.basename(sys.argv[1])[:14], "%Y%m%d%H%M%S")
    sat = os.path.basename(sys.argv[1])[15:22].replace("_", " ")
    print "***", reftime, sat

    with open(sys.argv[1]) as fp:
        lines = fp.read()

    hrpt = HRPT(sat, reftime)

    for truc in hrpt.read(lines):
        print truc[0][:4]
        # raw_input()

    try:
        serve(sys.argv[1])
    except KeyboardInterrupt:
        print "ok, stopping"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012, 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Setup Trollcast
"""

from setuptools import setup
import sys
import imp

version = imp.load_source('trollcast.version', 'trollcast/version.py')


requirements = ['pyzmq', 'pyorbital', 'posttroll',
                'numpy', 'pyinotify', 'trollsift']
if sys.version_info < (2, 7):
    requirements.append('argparse')


setup(name="trollcast",
      version=version.__version__,
      description='Weather satellite data exchange tool',
      author='The pytroll team',
      author_email='martin.raspaud@smhi.se',
      packages=['trollcast'],
      scripts=['bin/trollcast_client',
               'bin/trollcast_server'],
      zip_safe=False,
      license="GPLv3",
      install_requires=requirements,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python',
          'Operating System :: OS Independent',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
          'Topic :: Communications'
      ],
      #extras_require={ 'daemon': ['python-daemon'],}
      test_suite='trollcast.tests.suite',
      test_requires=['mock']
      )
